! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! Jan, 22th, 2015

! Forward Born modeling program, based on the velocity-stress
! elastic modeling program

! FD elastic operator module
! This is the main subroutine that calls
! the FD derivatives, sources, etc.
module born_mod

  use source_mod
  use seis_mod_b
  use bornlib_mod
  use snap_mod_b
  use pml_mod
  use omp_lib
  !use clib_mod

  implicit none
  integer, private :: nt, nx, nz, abc, zrec, ix
  integer, private :: xsource, zsource, nsource, nsnap, nseis
  real, private    :: dt, dx, dz, fp
  !real*16, private :: source
  real, private :: source
  logical, private :: free_surface

  contains

  ! Initialization of all variables needed by this subroutine
  subroutine elastic_init(nz_in,dz_in,nx_in,dx_in,nt_in,dt_in,xsource_in,  &
                          zsource_in,nsource_in, fp_in, nsnap_in, nseis_in,&
                          abc_in, zrec_in, free_surface_in)

    integer             :: nz_in,nx_in,nt_in,xsource_in,zsource_in,nsource_in
    integer             :: nsnap_in, nseis_in, abc_in, zrec_in
    real                :: dz_in,dx_in,dt_in, fp_in
    logical             :: free_surface_in

    nz = nz_in
    dz = dz_in
    nt = nt_in
    dt = dt_in
    nx = nx_in
    dx = dx_in
    fp = fp_in
    xsource = xsource_in
    zsource = zsource_in
    nsource = nsource_in
    nsnap   = nsnap_in
    nseis = nseis_in
    abc = abc_in
    zrec = zrec_in
    free_surface = free_surface_in

  end subroutine

  ! Function that is called by Main
  function elastic_op(vp0, vs0, rho0, dvp, dvs, drho, seis_Vx, seis_Vz, seis_P, &
                      seis_Vx_b, seis_Vz_b, seis_P_b, snap_Vx, snap_Vz, snap_P, &
                      snap_Vx_b, snap_Vz_b, snap_P_b, lame) result(stat)
    real,dimension(:,:)         :: vp0, vs0, rho0
    real,dimension(:,:)         :: dvp, dvs, drho
    real,dimension(:,:)         :: seis_Vx, seis_Vz, seis_P
    real,dimension(:,:,:)       :: snap_Vx, snap_Vz, snap_P
    real,dimension(:,:)         :: seis_Vx_b, seis_Vz_b, seis_P_b
    real,dimension(:,:,:)       :: snap_Vx_b, snap_Vz_b, snap_P_b
    integer                     :: stat
    logical,intent(in),optional :: lame

    stat=1
    call elastic_op2(vp0, vs0, rho0, dvp, dvs, drho, seis_Vx, seis_Vz, seis_P, &
                     seis_Vx_b, seis_Vz_b, seis_P_b, snap_Vx, snap_Vz, snap_P, &
                     snap_Vx_b, snap_Vz_b, snap_P_b, lame)
    stat=0
  end function

  ! Subroutine used by function
  subroutine elastic_op2(vp0, vs0, rho0, dvp, dvs, drho, seis_Vx, seis_Vz, seis_P, &
                         seis_Vx_b, seis_Vz_b, seis_P_b, snap_Vx, snap_Vz, snap_P, &
                         snap_Vx_b, snap_Vz_b, snap_P_b, lame)
    integer                             :: it, ix, iz, dsnap, dseis
    integer                             :: countsnap,countseis, temp
    real, dimension(nz,nx)              :: vp0,vs0,rho0
    real, dimension(nz,nx)              :: dvp,dvs,drho

    !Outputs
    real, dimension(nseis,nx-2*abc)     :: seis_Vx, seis_Vz, seis_P
    real, dimension(nz,nx,nsnap)        :: snap_Vx, snap_Vz, snap_P
    real, dimension(nseis,nx-2*abc)     :: seis_Vx_b, seis_Vz_b, seis_P_b
    real, dimension(nz,nx,nsnap)        :: snap_Vx_b, snap_Vz_b, snap_P_b

    !Background propagation
    real, allocatable, dimension(:,:)   :: mi0,l0,l2mi0
    real, allocatable, dimension(:,:)   :: Vxnew,Vznew,Vx_xnew,Vx_znew,Vz_xnew,Vz_znew
    real, allocatable, dimension(:,:)   :: sigmaxxnew,sigmazznew,sigmaxznew
    real, allocatable, dimension(:,:)   :: sigmaxx_xnew,sigmaxx_znew,sigmazz_xnew,sigmazz_znew
    real, allocatable, dimension(:,:)   :: sigmaxz_xnew,sigmaxz_znew
    real, allocatable, dimension(:,:)   :: Vxcur,Vzcur,Vx_xcur,Vx_zcur,Vz_xcur,Vz_zcur
    real, allocatable, dimension(:,:)   :: sigmaxxcur,sigmazzcur,sigmaxzcur
    real, allocatable, dimension(:,:)   :: sigmaxx_xcur,sigmaxx_zcur,sigmazz_xcur,sigmazz_zcur
    real, allocatable, dimension(:,:)   :: sigmaxz_xcur,sigmaxz_zcur
    real, allocatable, dimension(:,:)   :: Vxold,Vzold,Vx_xold,Vx_zold,Vz_xold,Vz_zold
    real, allocatable, dimension(:,:)   :: sigmaxxold,sigmazzold,sigmaxzold
    real, allocatable, dimension(:,:)   :: sigmaxx_xold,sigmaxx_zold,sigmazz_xold,sigmazz_zold
    real, allocatable, dimension(:,:)   :: sigmaxz_xold,sigmaxz_zold
    real, allocatable, dimension(:,:)   :: rho0_Vx,rho0_Vz
    real, allocatable, dimension(:,:)   :: drho_Vx,drho_Vz
    real                                :: aux1,t,nsource2

    !Scatterer propagation
    real, allocatable, dimension(:,:)   :: dmi, dl, dl2mi
    real, allocatable, dimension(:,:)   :: Vx_S, Vz_S
    real, allocatable, dimension(:,:)   :: sigmaxx_S, sigmazz_S, sigmaxz_S
    real, allocatable, dimension(:,:)   :: Vxnew_b,Vznew_b,Vx_xnew_b,Vx_znew_b,Vz_xnew_b,Vz_znew_b
    real, allocatable, dimension(:,:)   :: sigmaxxnew_b,sigmazznew_b,sigmaxznew_b
    real, allocatable, dimension(:,:)   :: sigmaxx_xnew_b,sigmaxx_znew_b,sigmazz_xnew_b,sigmazz_znew_b
    real, allocatable, dimension(:,:)   :: sigmaxz_xnew_b,sigmaxz_znew_b
    real, allocatable, dimension(:,:)   :: Vxcur_b,Vzcur_b,Vx_xcur_b,Vx_zcur_b,Vz_xcur_b,Vz_zcur_b
    real, allocatable, dimension(:,:)   :: sigmaxxcur_b,sigmazzcur_b,sigmaxzcur_b
    real, allocatable, dimension(:,:)   :: sigmaxx_xcur_b,sigmaxx_zcur_b,sigmazz_xcur_b,sigmazz_zcur_b
    real, allocatable, dimension(:,:)   :: sigmaxz_xcur_b,sigmaxz_zcur_b
    real, allocatable, dimension(:,:)   :: Vxold_b,Vzold_b,Vx_xold_b,Vx_zold_b,Vz_xold_b,Vz_zold_b
    real, allocatable, dimension(:,:)   :: sigmaxxold_b,sigmazzold_b,sigmaxzold_b
    real, allocatable, dimension(:,:)   :: sigmaxx_xold_b,sigmaxx_zold_b,sigmazz_xold_b,sigmazz_zold_b
    real, allocatable, dimension(:,:)   :: sigmaxz_xold_b,sigmaxz_zold_b
    integer                             :: stat
    logical,intent(in),optional         :: lame

    !Memory allocation

    !Allocate lame parameters
    allocate(mi0(nz,nx))
    allocate(l0(nz,nx))
    allocate(l2mi0(nz,nx))
    allocate(rho0_Vx(nz,nx))
    allocate(rho0_Vz(nz,nx))
    allocate(dmi(nz,nx))
    allocate(dl(nz,nx))
    allocate(dl2mi(nz,nx))
    allocate(drho_Vx(nz,nx))
    allocate(drho_Vz(nz,nx))

    !Allocate Velocities and stresses
    allocate(Vxnew(nz,nx))
    allocate(Vznew(nz,nx))
    allocate(Vx_xnew(nz,nx))
    allocate(Vx_znew(nz,nx))
    allocate(Vz_xnew(nz,nx))
    allocate(Vz_znew(nz,nx))
    allocate(sigmaxxnew(nz,nx))
    allocate(sigmazznew(nz,nx))
    allocate(sigmaxznew(nz,nx))
    allocate(sigmaxx_xnew(nz,nx))
    allocate(sigmaxx_znew(nz,nx))
    allocate(sigmazz_xnew(nz,nx))
    allocate(sigmazz_znew(nz,nx))
    allocate(sigmaxz_xnew(nz,nx))
    allocate(sigmaxz_znew(nz,nx))
    allocate(Vxcur(nz,nx))
    allocate(Vzcur(nz,nx))
    allocate(Vx_xcur(nz,nx))
    allocate(Vx_zcur(nz,nx))
    allocate(Vz_xcur(nz,nx))
    allocate(Vz_zcur(nz,nx))
    allocate(sigmaxxcur(nz,nx))
    allocate(sigmazzcur(nz,nx))
    allocate(sigmaxzcur(nz,nx))
    allocate(sigmaxx_xcur(nz,nx))
    allocate(sigmaxx_zcur(nz,nx))
    allocate(sigmazz_xcur(nz,nx))
    allocate(sigmazz_zcur(nz,nx))
    allocate(sigmaxz_xcur(nz,nx))
    allocate(sigmaxz_zcur(nz,nx))
    allocate(Vxold(nz,nx))
    allocate(Vzold(nz,nx))
    allocate(Vx_xold(nz,nx))
    allocate(Vx_zold(nz,nx))
    allocate(Vz_xold(nz,nx))
    allocate(Vz_zold(nz,nx))
    allocate(sigmaxxold(nz,nx))
    allocate(sigmazzold(nz,nx))
    allocate(sigmaxzold(nz,nx))
    allocate(sigmaxx_xold(nz,nx))
    allocate(sigmaxx_zold(nz,nx))
    allocate(sigmazz_xold(nz,nx))
    allocate(sigmazz_zold(nz,nx))
    allocate(sigmaxz_xold(nz,nx))
    allocate(sigmaxz_zold(nz,nx))

    !Born modeling propagation
    !Source wavefield
    allocate(Vx_S(nz,nx))
    allocate(Vz_S(nz,nx))
    allocate(sigmaxx_S(nz,nx))
    allocate(sigmazz_S(nz,nx))
    allocate(sigmaxz_S(nz,nx))
    !Scatterer wavefield
    allocate(Vxnew_b(nz,nx))
    allocate(Vznew_b(nz,nx))
    allocate(Vx_xnew_b(nz,nx))
    allocate(Vx_znew_b(nz,nx))
    allocate(Vz_xnew_b(nz,nx))
    allocate(Vz_znew_b(nz,nx))
    allocate(sigmaxxnew_b(nz,nx))
    allocate(sigmazznew_b(nz,nx))
    allocate(sigmaxznew_b(nz,nx))
    allocate(sigmaxx_xnew_b(nz,nx))
    allocate(sigmaxx_znew_b(nz,nx))
    allocate(sigmazz_xnew_b(nz,nx))
    allocate(sigmazz_znew_b(nz,nx))
    allocate(sigmaxz_xnew_b(nz,nx))
    allocate(sigmaxz_znew_b(nz,nx))
    allocate(Vxcur_b(nz,nx))
    allocate(Vzcur_b(nz,nx))
    allocate(Vx_xcur_b(nz,nx))
    allocate(Vx_zcur_b(nz,nx))
    allocate(Vz_xcur_b(nz,nx))
    allocate(Vz_zcur_b(nz,nx))
    allocate(sigmaxxcur_b(nz,nx))
    allocate(sigmazzcur_b(nz,nx))
    allocate(sigmaxzcur_b(nz,nx))
    allocate(sigmaxx_xcur_b(nz,nx))
    allocate(sigmaxx_zcur_b(nz,nx))
    allocate(sigmazz_xcur_b(nz,nx))
    allocate(sigmazz_zcur_b(nz,nx))
    allocate(sigmaxz_xcur_b(nz,nx))
    allocate(sigmaxz_zcur_b(nz,nx))
    allocate(Vxold_b(nz,nx))
    allocate(Vzold_b(nz,nx))
    allocate(Vx_xold_b(nz,nx))
    allocate(Vx_zold_b(nz,nx))
    allocate(Vz_xold_b(nz,nx))
    allocate(Vz_zold_b(nz,nx))
    allocate(sigmaxxold_b(nz,nx))
    allocate(sigmazzold_b(nz,nx))
    allocate(sigmaxzold_b(nz,nx))
    allocate(sigmaxx_xold_b(nz,nx))
    allocate(sigmaxx_zold_b(nz,nx))
    allocate(sigmazz_xold_b(nz,nx))
    allocate(sigmazz_zold_b(nz,nx))
    allocate(sigmaxz_xold_b(nz,nx))
    allocate(sigmaxz_zold_b(nz,nx))

    !Initialize values
    !Background properties
    mi0         =0.0
    l0          =0.0
    l2mi0       =0.0
    rho0_Vx     =0.0
    rho0_Vz     =0.0

    !Perturbation properties
    dmi         =0.0
    dl          =0.0
    dl2mi       =0.0
    drho_Vx     =0.0
    drho_Vz     =0.0

    !Background wavefield
    Vxnew         =0.0
    Vznew         =0.0
    Vx_xnew       =0.0
    Vx_znew       =0.0
    Vz_xnew       =0.0
    Vz_znew       =0.0
    sigmaxxnew    =0.0
    sigmazznew    =0.0
    sigmaxznew    =0.0
    sigmaxx_xnew  =0.0
    sigmaxx_znew  =0.0
    sigmazz_xnew  =0.0
    sigmazz_znew  =0.0
    sigmaxz_xnew  =0.0
    sigmaxz_znew  =0.0
    Vxcur         =0.0
    Vzcur         =0.0
    Vx_xcur       =0.0
    Vx_zcur       =0.0
    Vz_xcur       =0.0
    Vz_zcur       =0.0
    sigmaxxcur    =0.0
    sigmazzcur    =0.0
    sigmaxzcur    =0.0
    sigmaxx_xcur  =0.0
    sigmaxx_zcur  =0.0
    sigmazz_xcur  =0.0
    sigmazz_zcur  =0.0
    sigmaxz_xcur  =0.0
    sigmaxz_zcur  =0.0
    Vxold         =0.0
    Vzold         =0.0
    Vx_xold       =0.0
    Vx_zold       =0.0
    Vz_xold       =0.0
    Vz_zold       =0.0
    sigmaxxold    =0.0
    sigmazzold    =0.0
    sigmaxzold    =0.0
    sigmaxx_xold  =0.0
    sigmaxx_zold  =0.0
    sigmazz_xold  =0.0
    sigmazz_zold  =0.0
    sigmaxz_xold  =0.0
    sigmaxz_zold  =0.0

    !Born source
    Vx_S       =0.0
    Vz_S       =0.0
    sigmaxx_S  =0.0
    sigmazz_S  =0.0
    sigmaxz_S  =0.0

    !Scatterer wavefield
    Vxnew_b         =0.0
    Vznew_b         =0.0
    Vx_xnew_b       =0.0
    Vx_znew_b       =0.0
    Vz_xnew_b       =0.0
    Vz_znew_b       =0.0
    sigmaxxnew_b    =0.0
    sigmazznew_b    =0.0
    sigmaxznew_b    =0.0
    sigmaxx_xnew_b  =0.0
    sigmaxx_znew_b  =0.0
    sigmazz_xnew_b  =0.0
    sigmazz_znew_b  =0.0
    sigmaxz_xnew_b  =0.0
    sigmaxz_znew_b  =0.0
    Vxcur_b         =0.0
    Vzcur_b         =0.0
    Vx_xcur_b       =0.0
    Vx_zcur_b       =0.0
    Vz_xcur_b       =0.0
    Vz_zcur_b       =0.0
    sigmaxxcur_b    =0.0
    sigmazzcur_b    =0.0
    sigmaxzcur_b    =0.0
    sigmaxx_xcur_b  =0.0
    sigmaxx_zcur_b  =0.0
    sigmazz_xcur_b  =0.0
    sigmazz_zcur_b  =0.0
    sigmaxz_xcur_b  =0.0
    sigmaxz_zcur_b  =0.0
    Vxold_b         =0.0
    Vzold_b         =0.0
    Vx_xold_b       =0.0
    Vx_zold_b       =0.0
    Vz_xold_b       =0.0
    Vz_zold_b       =0.0
    sigmaxxold_b    =0.0
    sigmazzold_b    =0.0
    sigmaxzold_b    =0.0
    sigmaxx_xold_b  =0.0
    sigmaxx_zold_b  =0.0
    sigmazz_xold_b  =0.0
    sigmazz_zold_b  =0.0
    sigmaxz_xold_b  =0.0
    sigmaxz_zold_b  =0.0

    dseis      =(nt/nseis)+1
    dsnap      =nt/nsnap
    countsnap  =1
    countseis  =1
    nsource2   =2*nint(sqrt(6.0)/3.14159265/fp/dt)

    !Define Lame parameters
    !These model properties are still in the original grid, centered at i,j
    do ix=1,nx
      do iz=1,nz
        if(lame)then
          l0(iz,ix)=vp0(iz,ix)
          mi0(iz,ix)=vs0(iz,ix)

          dl(iz,ix)=dvp(iz,ix)
          dmi(iz,ix)=dvs(iz,ix)
        else
          mi0(iz,ix)=vs0(iz,ix)*vs0(iz,ix)*rho0(iz,ix)
          l0(iz,ix)=vp0(iz,ix)*vp0(iz,ix)*rho0(iz,ix)-2.0*mi0(iz,ix)
          l2mi0(iz,ix)=l0(iz,ix)+2.0*mi0(iz,ix)

          dmi(iz,ix)=dvs(iz,ix)*dvs(iz,ix)*drho(iz,ix)
          dl(iz,ix)=dvp(iz,ix)*dvp(iz,ix)*drho(iz,ix)-2.0*dmi(iz,ix)
          dl2mi(iz,ix)=dl(iz,ix)+2.0*dmi(iz,ix)
        endif
      enddo
    enddo

    !Calculate inverse of density
    !These also calculate the values of density at the velocity staggered grid,
    !using the arithmetic average proposed in Moczo, 2002
    !Rho_Vx

    aux1=(dt/dx)

    !Rho_vx
    do ix=1,nx
      do iz=1,nz-1
        rho0_Vx(iz,ix)=aux1*1.0/(0.5*(rho0(iz+1,ix) + rho0(iz,ix)))
        drho_Vx(iz,ix)=-(drho(iz+1,ix)/rho0(iz+1,ix) +&
                         drho(iz  ,ix)/rho0(iz  ,ix))
      enddo
    enddo
    !Rho_vz
    do ix=1,nx-1
      do iz=1,nz
        rho0_Vz(iz,ix)=aux1*1.0/(0.5*(rho0(iz,ix+1) + rho0(iz,ix)))
        drho_Vz(iz,ix)=-(drho(iz,ix+1)/rho0(iz,ix+1) +&
                         drho(iz,ix  )/rho0(iz,ix  ))
      enddo
    enddo
    !Right side
    do iz=1,nz
      rho0_Vz(iz,nx)=aux1*(1.0/rho0(iz,nx))
      drho_Vz(iz,nx)=-drho(iz,nx)/rho0(iz,nx)
    enddo
    !Bottom
    do ix=1,nx
      rho0_Vx(nz,ix)=aux1*(1.0/rho0(nz,ix))
      drho_Vx(nz,ix)=-drho(nz,ix)/rho0(nz,ix)
    enddo
    
    !Calculate weigthed average for the lambda variable
    !These put lambda and (lambda+2mi) in their respective staggered grids,
    !using the harmonic average proposed in Moczo, 2002
    do ix=2,nx-1
      do iz=2,nz-1
        l0(iz,ix)=aux1*4.0/((1.0/l0(iz,ix))+(1.0/l0(iz+1,ix))+(1.0/l0(iz,ix+1))+(1.0/l0(iz+1,ix+1)))
        if (mi0(iz,ix)==0.or.mi0(iz+1,ix)==0.or.mi0(iz,ix+1)==0.or.mi0(iz+1,ix+1)==0) then
          mi0(iz,ix)=0.0
        else
          mi0(iz,ix)=aux1*4.0/((1.0/mi0(iz,ix))+(1.0/mi0(iz+1,ix))+(1.0/mi0(iz,ix+1))+(1.0/mi0(iz+1,ix+1)))
        endif
        if (dl(iz,ix)<1e-2.or.dl(iz+1,ix)<1e-2.or.dl(iz,ix+1)<1e-2.or.dl(iz+1,ix+1)<1e-2) then
          dl(iz,ix)=0.0
        else
          dl(iz,ix)=aux1*4.0/((1.0/dl(iz,ix))+(1.0/dl(iz+1,ix))+(1.0/dl(iz,ix+1))+(1.0/dl(iz+1,ix+1)))
        endif
        if (dmi(iz,ix)<1e-2.or.dmi(iz+1,ix)<1e-2.or.dmi(iz,ix+1)<1e-2.or.dmi(iz+1,ix+1)<1e-2) then
          dmi(iz,ix)=0.0
        else
          dmi(iz,ix)=aux1*4.0/((1.0/dmi(iz,ix))+(1.0/dmi(iz+1,ix))+(1.0/dmi(iz,ix+1))+(1.0/dmi(iz+1,ix+1)))
        endif
        l2mi0(iz,ix)=l0(iz,ix)+2.0*mi0(iz,ix)
        dl2mi(iz,ix)=dl(iz,ix)+2.0*dmi(iz,ix)
      enddo
    enddo

    !Initialize subroutines
    call source_init(fp,dt,dx,nsource,nz,nx)
    call seis_init(nz,nx, abc, zrec, seis_Vx, seis_Vz, seis_P, seis_Vx_b, seis_Vz_b, seis_P_b)
    call bornmod_init(nz,nx,rho0_Vx,rho0_Vz,l0,mi0,l2mi0,drho_vx,drho_vz,dl,dmi,dl2mi)
    call pml_init(nz,nx,dx,dt,abc,vp0,free_surface)
    call snap_init(nz,nx,abc,snap_Vx,snap_Vz, snap_P, snap_Vx_b, snap_Vz_b, snap_P_b)
    
    !Time loop for 2nd order time
    write(0,*)"============================================================="
    write(0,*)"BEGIN OF TIME MODELING"
    write(0,*)"============================================================="
    write(0,*)"Processing:"
    write(0,*)"0%           25%            50%           75%            100%"
    write(0,'(a)',advance="no")"["
    do it=1,nt
      if (it<nsource) then
        call source_op2(it, source)
      else
        source=0.0
      endif
      source = 0.7E7*source*(dt/(dx*dx))*0.25

!###################################################
! Propagation of the wavefields in the background
!###################################################
      !Vx_x
      call Ax(.false.,.false.,sigmaxx_xcur,Vx_xnew)
      call Ax(.false.,.true.,sigmaxx_zcur,Vx_xnew)
      Vx_xnew = Vx_xnew + Vx_xold
      !call L(.false.,.true.,Vx_xold,Vx_xnew)
      !call explosive_source(.false.,.false.,source,xsource,zsource,temp1)
      !call C(.false.,.true.,temp1,Vx_xnew)
      !call K(.false.,.true.,Vx_xnew,Vx_xnew)

      !Vx_z
      call Bx(.false.,.false.,sigmaxz_xcur,Vx_znew)
      call Bx(.false.,.true.,sigmaxz_zcur,Vx_znew)
      Vx_znew = Vx_znew + Vx_zold
      !call N(.false.,.true.,Vx_zold,Vx_znew)
      !call explosive_source(.false.,.false.,source,xsource,zsource,temp1)
      !call C(.false.,.true.,temp1,Vx_znew)
      !call M(.false.,.true.,Vx_znew,Vx_znew)

      !Vz_x
      call Az(.false.,.false.,sigmaxz_xcur,Vz_xnew)
      call Az(.false.,.true.,sigmaxz_zcur,Vz_xnew)
      Vz_xnew = Vz_xnew + Vz_xold
      !call P(.false.,.true.,Vz_xold,Vz_xnew)
      !call explosive_source(.false.,.false.,source,xsource,zsource,temp1)
      !call C(.false.,.true.,temp1,Vz_xnew)
      !call O(.false.,.true.,Vz_xnew,Vz_xnew)

      !Vz_z
      call Bz(.false.,.false.,sigmazz_xcur,Vz_znew)
      call Bz(.false.,.true.,sigmazz_zcur,Vz_znew)
      Vz_znew = Vz_znew + Vz_zold
      !call R(.false.,.true.,Vz_zold,Vz_znew)
      !call explosive_source(.false.,.false.,source,xsource,zsource,temp1)
      !call C(.false.,.true.,temp1,Vz_znew)
      !call Q(.false.,.true.,Vz_znew,Vz_znew)

      !sigmaxx_x
      call D(.false.,.false.,Vx_xcur,sigmaxx_xnew)
      call D(.false.,.true.,Vx_zcur,sigmaxx_xnew)
      sigmaxx_xnew = sigmaxx_xnew + sigmaxx_xold
      !call P(.false.,.true.,sigmaxx_xold,sigmaxx_xnew)
      call explosive_sourcex(.false.,.true.,source,xsource,zsource,sigmaxx_xnew)
      !call O(.false.,.true.,sigmaxx_xnew,sigmaxx_xnew)

      !sigmaxx_z
      call E(.false.,.false.,Vz_xcur,sigmaxx_znew)
      call E(.false.,.true.,Vz_zcur,sigmaxx_znew)
      sigmaxx_znew = sigmaxx_znew + sigmaxx_zold
      !call R(.false.,.true.,sigmaxx_zold,sigmaxx_znew)
      call explosive_sourcez(.false.,.true.,source,xsource,zsource,sigmaxx_znew)
      !call Q(.false.,.true.,sigmaxx_znew,sigmaxx_znew)

      !sigmazz_x
      call F(.false.,.false.,Vx_xcur,sigmazz_xnew)
      call F(.false.,.true.,Vx_zcur,sigmazz_xnew)
      sigmazz_xnew = sigmazz_xnew + sigmazz_xold
      !call P(.false.,.true.,sigmazz_xold,sigmazz_xnew)
      call explosive_sourcex(.false.,.true.,source,xsource,zsource,sigmazz_xnew)
      !call O(.false.,.true.,sigmazz_xnew,sigmazz_xnew)

      !sigmazz_z
      call G(.false.,.false.,Vz_xcur,sigmazz_znew)
      call G(.false.,.true.,Vz_zcur,sigmazz_znew)
      sigmazz_znew = sigmazz_znew + sigmazz_zold
      !call R(.false.,.true.,sigmazz_zold,sigmazz_znew)
      call explosive_sourcez(.false.,.true.,source,xsource,zsource,sigmazz_znew)
      !call Q(.false.,.true.,sigmaxx_znew,sigmaxx_znew)

      !sigmaxz_x
      call H(.false.,.false.,Vx_xcur,sigmaxz_xnew)
      call H(.false.,.true.,Vx_zcur,sigmaxz_xnew)
      sigmaxz_xnew = sigmaxz_xnew + sigmaxz_xold
      !call L(.false.,.true.,sigmaxz_xold,sigmaxz_xnew)
      !call explosive_source(.false.,.true.,source,xsource,zsource,sigmaxz_xnew)
      !call K(.false.,.true.,sigmazz_xnew,sigmazz_xnew)

      !sigmaxz_z
      call J(.false.,.false.,Vz_xcur,sigmaxz_znew)
      call J(.false.,.true.,Vz_zcur,sigmaxz_znew)
      sigmaxz_znew = sigmaxz_znew + sigmaxz_zold
      !call N(.false.,.true.,sigmaxz_zold,sigmaxz_znew)
      !call explosive_source(.false.,.true.,source,xsource,zsource,sigmaxz_znew)
      !call M(.false.,.true.,sigmaxx_znew,sigmaxx_znew)

!###################################################
! Propagation of the scattered wavefield
!###################################################
      !Vx_x
      call Ax(.false.,.false.,sigmaxx_xcur_b,Vx_xnew_b)
      call Ax(.false.,.true.,sigmaxx_zcur_b,Vx_xnew_b)
      Vx_xnew_b = Vx_xnew_b + Vx_xold_b
      !call L(.false.,.true.,Vx_xold_b,Vx_xnew_b)
      call S1(.false.,.true.,Vx_xnew,Vx_xcur,Vx_xnew_b)
      !call C(.false.,.true.,temp1,Vx_xnew_b)
      !call K(.false.,.true.,Vx_xnew_b,Vx_xnew_b)

      !Vx_z
      call Bx(.false.,.false.,sigmaxz_xcur_b,Vx_znew_b)
      call Bx(.false.,.true.,sigmaxz_zcur_b,Vx_znew_b)
      Vx_znew_b = Vx_znew_b + Vx_zold_b
      !call N(.false.,.true.,Vx_zold_b,Vx_znew_b)
      call S1(.false.,.true.,Vx_znew,Vx_zcur,Vx_znew_b)
      !call C(.false.,.true.,temp1,Vx_znew_b)
      !call M(.false.,.true.,Vx_znew_b,Vx_znew_b)

      !Vz_x
      call Az(.false.,.false.,sigmaxz_xcur_b,Vz_xnew_b)
      call Az(.false.,.true.,sigmaxz_zcur_b,Vz_xnew_b)
      Vz_xnew_b = Vz_xnew_b + Vz_xold_b
      !call P(.false.,.true.,Vz_xold_b,Vz_xnew_b)
      call S2(.false.,.true.,Vz_xnew,Vz_xcur,Vz_xnew_b)
      !call C(.false.,.true.,temp1,Vz_xnew_b)
      !call O(.false.,.true.,Vz_xnew_b,Vz_xnew_b)

      !Vz_z
      call Bz(.false.,.false.,sigmazz_xcur_b,Vz_znew_b)
      call Bz(.false.,.true.,sigmazz_zcur_b,Vz_znew_b)
      Vz_znew_b = Vz_znew_b + Vz_zold_b
      !call R(.false.,.true.,Vz_zold_b,Vz_znew_b)
      call S2(.false.,.true.,Vz_znew,Vz_zcur,Vz_znew_b)
      !call C(.false.,.true.,temp1,Vz_znew_b)
      !call Q(.false.,.true.,Vz_znew_b,Vz_znew_b)

      !sigmaxx_x
      call D(.false.,.false.,Vx_xcur_b,sigmaxx_xnew_b)
      call D(.false.,.true.,Vx_zcur_b,sigmaxx_xnew_b)
      sigmaxx_xnew_b = sigmaxx_xnew_b + sigmaxx_xold_b
      !call P(.false.,.true.,sigmaxx_xold_b,sigmaxx_xnew_b)
      call S3(.false.,.true.,Vx_xnew,sigmaxx_xnew_b)
      call S3(.false.,.true.,Vx_znew,sigmaxx_xnew_b)
      !call O(.false.,.true.,sigmaxx_xnew_b,sigmaxx_xnew_b)

      !sigmaxx_z
      call E(.false.,.false.,Vz_xcur_b,sigmaxx_znew_b)
      call E(.false.,.true.,Vz_zcur_b,sigmaxx_znew_b)
      sigmaxx_znew_b = sigmaxx_znew_b + sigmaxx_zold_b
      !call R(.false.,.true.,sigmaxx_zold_b,sigmaxx_znew_b)
      call S4(.false.,.true.,Vz_xnew,sigmaxx_znew_b)
      call S4(.false.,.true.,Vz_znew,sigmaxx_znew_b)
      !call Q(.false.,.true.,sigmaxx_znew_b,sigmaxx_znew_b)

      !sigmazz_x
      call F(.false.,.false.,Vx_xcur_b,sigmazz_xnew_b)
      call F(.false.,.true.,Vx_zcur_b,sigmazz_xnew_b)
      sigmazz_xnew_b = sigmazz_xnew_b + sigmazz_xold_b
      !call P(.false.,.true.,sigmazz_xold_b,sigmazz_xnew_b)
      call S5(.false.,.true.,Vx_xnew,sigmazz_xnew_b)
      call S5(.false.,.true.,Vx_znew,sigmazz_xnew_b)
      !call O(.false.,.true.,sigmazz_xnew_b,sigmazz_xnew_b)

      !sigmazz_z
      call G(.false.,.false.,Vz_xcur_b,sigmazz_znew_b)
      call G(.false.,.true.,Vz_zcur_b,sigmazz_znew_b)
      sigmazz_znew_b = sigmazz_znew_b + sigmazz_zold_b
      !call R(.false.,.true.,sigmazz_zold_b,sigmazz_znew_b)
      call S6(.false.,.true.,Vz_xnew,sigmazz_znew_b)
      call S6(.false.,.true.,Vz_znew,sigmazz_znew_b)
      !call Q(.false.,.true.,sigmaxx_znew_b,sigmaxx_znew_b)

      !sigmaxz_x
      call H(.false.,.false.,Vx_xcur_b,sigmaxz_xnew_b)
      call H(.false.,.true.,Vx_zcur_b,sigmaxz_xnew_b)
      sigmaxz_xnew_b = sigmaxz_xnew_b + sigmaxz_xold_b
      !call L(.false.,.true.,sigmaxz_xold_b,sigmaxz_xnew_b)
      call S7(.false.,.true.,Vx_xnew,sigmaxz_xnew_b)
      call S7(.false.,.true.,Vx_znew,sigmaxz_xnew_b)
      !call K(.false.,.true.,sigmazz_xnew_b,sigmazz_xnew_b)

      !sigmaxz_z
      call J(.false.,.false.,Vz_xcur_b,sigmaxz_znew_b)
      call J(.false.,.true.,Vz_zcur_b,sigmaxz_znew_b)
      sigmaxz_znew_b = sigmaxz_znew_b + sigmaxz_zold_b
      !call N(.false.,.true.,sigmaxz_zold_b,sigmaxz_znew_b)
      call S8(.false.,.true.,Vz_xnew,sigmaxz_znew_b)
      call S8(.false.,.true.,Vz_znew,sigmaxz_znew_b)
      !call M(.false.,.true.,sigmaxx_znew_b,sigmaxx_znew_b)

!###################################################
! Outputs and time updates
!###################################################
      Vxnew = Vx_xnew + Vx_znew
      Vznew = Vz_xnew + Vz_znew
      sigmaxxnew = sigmaxx_xnew + sigmaxx_znew
      sigmazznew = sigmazz_xnew + sigmazz_znew
      Vxnew_b = Vx_xnew_b + Vx_znew_b
      Vznew_b = Vz_xnew_b + Vz_znew_b
      sigmaxxnew_b = sigmaxx_xnew_b + sigmaxx_znew_b
      sigmazznew_b = sigmazz_xnew_b + sigmazz_znew_b

      !Write seismograms
      if (it.gt.nsource2.and.mod(it,dseis)==0) then
        call seis_op2(countseis, Vx_xnew, Vx_znew, Vz_xnew,Vz_znew)
        call seis_op3(countseis, sigmaxx_xnew, sigmaxx_znew, sigmazz_xnew, sigmazz_znew)
        call seis_op4(countseis, Vx_xnew_b, Vx_znew_b, Vz_xnew_b, Vz_znew_b)
        call seis_op5(countseis, sigmaxx_xnew_b, sigmaxx_znew_b, sigmazz_xnew_b, sigmazz_znew_b)
        countseis = countseis + 1
        !write(0,*)"Seismogram time sample"
      endif

      !Write snapshots
      if (mod(it,dsnap)==0.and.countsnap.le.nsnap) then
        call snap_op2(countsnap,Vx_xnew,Vx_znew,Vz_xnew,Vz_znew,Vx_xnew_b,Vx_znew_b,Vz_xnew_b,Vz_znew_b)
        call snap_op3(countsnap,sigmaxx_xnew,sigmaxx_znew,sigmazz_xnew,sigmazz_znew,sigmaxx_xnew_b,sigmaxx_znew_b,sigmazz_xnew_b,sigmazz_znew_b)
        countsnap = countsnap + 1
      endif

      Vx_xold = Vx_xcur
      Vx_zold = Vx_zcur
      Vz_xold = Vz_xcur
      Vz_zold = Vz_zcur
      sigmaxx_xold = sigmaxx_xcur
      sigmaxx_zold = sigmaxx_zcur
      sigmazz_xold = sigmazz_xcur
      sigmazz_zold = sigmazz_zcur
      sigmaxz_xold = sigmaxz_xcur
      sigmaxz_zold = sigmaxz_zcur

      Vx_xcur = Vx_xnew
      Vx_zcur = Vx_znew
      Vz_xcur = Vz_xnew
      Vz_zcur = Vz_znew
      sigmaxx_xcur = sigmaxx_xnew
      sigmaxx_zcur = sigmaxx_znew
      sigmazz_xcur = sigmazz_xnew
      sigmazz_zcur = sigmazz_znew
      sigmaxz_xcur = sigmaxz_xnew
      sigmaxz_zcur = sigmaxz_znew

      Vx_xold_b = Vx_xcur_b
      Vx_zold_b = Vx_zcur_b
      Vz_xold_b = Vz_xcur_b
      Vz_zold_b = Vz_zcur_b
      sigmaxx_xold_b = sigmaxx_xcur_b
      sigmaxx_zold_b = sigmaxx_zcur_b
      sigmazz_xold_b = sigmazz_xcur_b
      sigmazz_zold_b = sigmazz_zcur_b
      sigmaxz_xold_b = sigmaxz_xcur_b
      sigmaxz_zold_b = sigmaxz_zcur_b

      Vx_xcur_b = Vx_xnew_b
      Vx_zcur_b = Vx_znew_b
      Vz_xcur_b = Vz_xnew_b
      Vz_zcur_b = Vz_znew_b
      sigmaxx_xcur_b = sigmaxx_xnew_b
      sigmaxx_zcur_b = sigmaxx_znew_b
      sigmazz_xcur_b = sigmazz_xnew_b
      sigmazz_zcur_b = sigmazz_znew_b
      sigmaxz_xcur_b = sigmaxz_xnew_b
      sigmaxz_zcur_b = sigmaxz_znew_b

      !Write to terminal
      if (mod(it,(nt/59))==0) then
        write(0,'(a)',advance="no")"#"
        close(0)
      endif
    enddo
    write(0,'(a)',advance="yes")"]"
    write(0,*)"Last seismogram sample is:",countseis

  end subroutine

end module
